#include "customgraphicsitemdemo.h"
#include "ui_customgraphicsitemdemo.h"

#include <QDebug>

#include <QGraphicsPixmapItem>
#include <QGraphicsSceneMouseEvent>

const int AreaSelector::RESIZE_ZONE_SIZE = 20;
const int AreaSelector::MIN_AREA_SIZE = 100;

AreaSelector::AreaSelector( const QRectF& initialRect ) :
    m_innerRect( initialRect ), m_hovered( false ), m_activeZone( AreaSelector::NONE ) {

    setAcceptHoverEvents( true );
}

AreaSelector::~AreaSelector() { }

QRectF AreaSelector::boundingRect() const {
    return QRectF(
               m_innerRect.left() - RESIZE_ZONE_SIZE / 2.0,
               m_innerRect.top() - RESIZE_ZONE_SIZE / 2.0,
               m_innerRect.width() + RESIZE_ZONE_SIZE,
               m_innerRect.height() + RESIZE_ZONE_SIZE
           );
}

void AreaSelector::paint( QPainter* painter, const QStyleOptionGraphicsItem*, QWidget* ) {
    painter->drawRect( m_innerRect );

    if( m_hovered ) {
        QHash< SelectorZone, QRectF > zones = generateZones();
        foreach( const SelectorZone& key, zones.keys() ) {
            painter->fillRect( zones[ key ], Qt::white );
            painter->drawRect( zones[ key ] );
        }
    }
}

void AreaSelector::hoverEnterEvent( QGraphicsSceneHoverEvent* ) {
    m_hovered = true;
    update();

}

void AreaSelector::hoverLeaveEvent( QGraphicsSceneHoverEvent* ) {
    m_hovered = false;
    update();
}

void AreaSelector::mousePressEvent( QGraphicsSceneMouseEvent* event ) {
    switch( event->button() ) {
    case Qt::LeftButton: {
        QHash< SelectorZone, QRectF > zones = generateZones();
        foreach( const SelectorZone& key, zones.keys() ) {
            if( zones[ key ].contains( event->pos() ) ) {
                m_activeZone = key;
                break;
            }
        }

        if(
            m_activeZone == AreaSelector::NONE &&
            m_innerRect.contains( event->pos() )
        ) {
            m_activeZone = MOVE_ZONE;
            m_centerOffset = event->pos() - m_innerRect.center();
        }

        break;
    }

    default:
        // Nothing to do here
        break;
    }
}

void AreaSelector::mouseMoveEvent( QGraphicsSceneMouseEvent* event ) {
    prepareGeometryChange();

    QRectF sRect;

    if( scene() ) {
        sRect = scene()->sceneRect();
    }


    switch( m_activeZone ) {
    case MOVE_ZONE:
        m_innerRect.moveCenter( event->pos() - m_centerOffset );

        if( m_innerRect.left() < sRect.left() ) {
            m_innerRect.moveLeft( sRect.left() );
        }

        if( m_innerRect.top() < sRect.top() ) {
            m_innerRect.moveTop( sRect.top() );
        }

        if( sRect.right() < m_innerRect.right() ) {
            m_innerRect.moveRight( sRect.right() );
        }

        if( sRect.bottom() < m_innerRect.bottom() ) {
            m_innerRect.moveBottom( sRect.bottom() );
        }

        break;

    case RESIZE_TOP_LEFT_ZONE:
        m_innerRect.setTopLeft( event->pos() );
        break;

    case RESIZE_TOP_RIGHT_ZONE:
        m_innerRect.setTopRight( event->pos() );
        break;

    case RESIZE_BOTTOM_LEFT_ZONE:
        m_innerRect.setBottomLeft( event->pos() );
        break;

    case RESIZE_BOTTOM_RIGHT_ZONE:
        m_innerRect.setBottomRight( event->pos() );
        break;

    default:
        // Nothing to do here
        break;

    }

    if( m_activeZone != AreaSelector::NONE ) {
        if( m_innerRect.left() < sRect.left() ) {
            m_innerRect.setLeft( sRect.left() );
        }

        if( m_innerRect.top() < sRect.top() ) {
            m_innerRect.setTop( sRect.top() );
        }

        if( sRect.right() < m_innerRect.right() ) {
            m_innerRect.setRight( sRect.right() );
        }

        if( sRect.bottom() < m_innerRect.bottom() ) {
            m_innerRect.setBottom( sRect.bottom() );
        }

        if( m_innerRect.width() < MIN_AREA_SIZE ) {
            if( m_activeZone == RESIZE_BOTTOM_LEFT_ZONE || m_activeZone == RESIZE_TOP_LEFT_ZONE ) {
                m_innerRect.setLeft( m_innerRect.right() - MIN_AREA_SIZE );
            } else {
                m_innerRect.setRight( m_innerRect.left() + MIN_AREA_SIZE );
            }
        }

        if( m_innerRect.height() < MIN_AREA_SIZE ) {
            if( m_activeZone == RESIZE_TOP_LEFT_ZONE || m_activeZone == RESIZE_TOP_RIGHT_ZONE ) {
                m_innerRect.setTop( m_innerRect.bottom() - MIN_AREA_SIZE );
            } else {
                m_innerRect.setBottom( m_innerRect.top() + MIN_AREA_SIZE );
            }
        }
    }

    emit selectedAreaChanged( m_innerRect );
}

void AreaSelector::mouseReleaseEvent( QGraphicsSceneMouseEvent* ) {
    m_activeZone = AreaSelector::NONE;
}

QHash< AreaSelector::SelectorZone, QRectF > AreaSelector::generateZones() const {
    QRectF zoneRect( 0, 0, RESIZE_ZONE_SIZE, RESIZE_ZONE_SIZE );
    QHash< SelectorZone, QPointF > zoneCenters = QHash< SelectorZone, QPointF >();
    zoneCenters[ RESIZE_TOP_LEFT_ZONE ] = m_innerRect.topLeft();
    zoneCenters[ RESIZE_TOP_RIGHT_ZONE ] = m_innerRect.topRight();
    zoneCenters[ RESIZE_BOTTOM_LEFT_ZONE ] = m_innerRect.bottomLeft();
    zoneCenters[ RESIZE_BOTTOM_RIGHT_ZONE ] = m_innerRect.bottomRight();

    QHash< SelectorZone, QRectF > zones;
    foreach( const SelectorZone& key, zoneCenters.keys() ) {
        zoneRect.moveCenter( zoneCenters[ key ] );
        zones[ key ] = zoneRect;
    }

    return zones;
}

// ********************************************************************************

CustomGraphicsItemDemo::CustomGraphicsItemDemo( QWidget* parent ) :
    QWidget( parent ),
    ui( new Ui::CustomGraphicsItemDemo ) {
    ui->setupUi( this );

    ui->grView->setScene( &m_scene );

    m_pix.load( ":/images/image.jpeg" );
    m_scene.setSceneRect( 0, 0, m_pix.width(), m_pix.height() );
    m_scene.addPixmap( m_pix );

    QRectF initialRect( 0, 0, m_pix.width() * 0.2f, m_pix.height() * 0.2f );
    initialRect.moveCenter( m_scene.sceneRect().center() );
    AreaSelector* areaSelector = new AreaSelector( initialRect );
    connect( areaSelector, SIGNAL( selectedAreaChanged( QRectF ) ), SLOT( onSelectedAreaChanged( QRectF ) ) );
    m_scene.addItem( areaSelector );

    resizeEvent( NULL );
}

CustomGraphicsItemDemo::~CustomGraphicsItemDemo() {
    delete ui;
}

void CustomGraphicsItemDemo::resizeEvent( QResizeEvent* ) {
    static const int GRAPHICS_VIEW_MARGIN = 20;
    double scale = std::min(
                       ( ui->grView->width() - GRAPHICS_VIEW_MARGIN ) / m_scene.width(),
                       ( ui->grView->height() - GRAPHICS_VIEW_MARGIN ) / m_scene.height()
                   );

    ui->grView->resetMatrix();
    ui->grView->scale( scale, scale );
}

void CustomGraphicsItemDemo::onSelectedAreaChanged( const QRectF& area ) {
    static const int SCALED_VIEW_MARGIN = 20;
    ui->lbScaledView->setPixmap(
        m_pix.copy( area.toRect() ).scaled(
            ui->lbScaledView->size().width() - SCALED_VIEW_MARGIN,
            ui->lbScaledView->size().height() - SCALED_VIEW_MARGIN,
            Qt::KeepAspectRatio
        )
    );
}
